#!/usr/bin/python3

# Copyright (C) 2012-2013 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Lightweight command line client for linknx. It is aimed at reading or writing object values from/to linknx.
When reading objects, this script outputs a line per object, each line composed of the object id and its corresponding value separated by spaces.

SYNTAX:
        pyknxclient.py [-h host] [-p port] [-v level] -r id1 [-r id2 [...]]
        pyknxclient.py [-h host] [-p port] [-v level] -R pattern
        pyknxclient.py [-h host] [-p port] [-v level] -w id value

        Where id, id1, id2, ... are object identifiers specified in the linknx configuration XML with the 'id' attribute.

OPTIONS:
        -h, --host      Hostname of the machine running the linknx daemon (default is 'localhost').
        -p, --port      Port linknx listens on (default is 1028).
        -v, --verbose   Level of verbosity. Value must be one of the logging module (error, warning, info, debug)
        -r, --read      Read value of object with given id. Can occur multiple times with various identifiers.
        -R, --regex     Read value of all objects whose identifiers match the given regex pattern. The pattern must comply with the 're' python module.
        -w, --write     Writes a new value to the object of given identifier.
            --help      Display this help message and quit.

EXAMPLES:
        Read all objects:
            pyknxclient.py -R ".*"

        Read two objects:
            pyknxclient.py -r KitchenLights -r LivingRoomLights

        Turn off kitchen lights:
            pyknxclient.py -w KitchenLights off
            pyknxclient.py -w KitchenLights false
            pyknxclient.py -w KitchenLights 0
"""

import sys
import getopt
import traceback
import logging
from xml.dom.minidom import parseString
from threading import *
from pyknx import logger
from pyknx.linknx import *

def printUsage():
    print(__doc__)

if __name__ == '__main__':
    logger.initLogger(None, logging.INFO, usesDetailedLogging=False)

    try:
        options, remainder = getopt.getopt(sys.argv[1:], 'r:w:h:p:v:R:', ['read=', 'write=', 'regex=', 'host=', 'port=','verbose=','help'])
    except getopt.GetoptError as err:
        logger.reportError(sys.exc_info()[1])
        sys.exit(2)

    print('*******************')
    print('DEPRECATION NOTICE:')
    print('*******************')
    print('This script is now deprecated and its functionalities have been split into the two scripts pyknxread.py and pyknxwrite.py. An additional script named pyknxexecute.py allows for executing actions (which functionality is not handled by pyknxclient.py at all).')

    # Parse command line arguments.
    reads = False
    writes = False
    objectIds = []
    isRegex = False
    host = 'localhost'
    port = 1028
    verbosity = logging.WARNING
    for option, value in options:
        if option == '-r' or option == '--read':
            reads = True
            objectIds.append(value)
        elif option == '-R' or option == '--regex':
            reads = True
            objectIds.append(value)
            isRegex = True
        elif option == '-w' or option == '--write':
            writes = True
            objectIds.append(value)
        elif option == '-h' or option == '--host':
            host = value
        elif option == '-p' or option == '--port':
            port = value
        elif option == '-v' or option == '--verbose':
            verbosity = logger.parseLevel(value)
        elif option == '--help':
            printUsage()
            sys.exit(1)
        else:
            print('Unrecognized option ' + option)
            sys.exit(2)

    logger.initLogger(None, verbosity, usesDetailedLogging=False)

    if reads == writes:
        logger.reportError('Expecting -r|--read or -w|--write and not both.')
        sys.exit(2)

    if writes and len(objectIds) > 1:
        logger.reportError('Can only write one object.')
        sys.exit(2)

    valueToWrite = None
    if writes:
        if not remainder:
            print('No value specified.')
            printUsage()
            sys.exit(3)
        valueToWrite = remainder[0]
        del remainder[0]

    if remainder:
        logger.reportError('Too many arguments: ' + str(remainder))
        sys.exit(4)

    # Start linknx.
    linknx = Linknx(host, int(port))
    try:
        if reads:
            report={}
            if not isRegex:
                for objId in objectIds:
                    report[objId] = linknx.getObject(objId).value
            else:
                for obj in linknx.getObjects(objectIds[0]):
                    report[obj.id] = obj.value

            # No object.
            if not report:
                logger.reportWarning('No object of given id.')
                sys.exit(10)

            # Count tabs to align columns.
            longestId = max([len(id) for id in report.keys()])
            for o, v in report.items():
                spaceCount = longestId - len(o)
                spaces=''
                while spaceCount > 0:
                    spaces+=' '
                    spaceCount -= 1
                print('{0} {2} {1}'.format(o, v, spaces))
        elif writes:
            linknx.getObject(objectIds[0]).value = valueToWrite
    except Exception as e:
        if verbosity == logging.DEBUG:
            logger.reportException()
        else:
            logger.reportError(sys.exc_info()[1])
        sys.exit(3)



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Outputs the current version of the Pyknx package.
"""

import pyknx
import argparse
import sys
import logging
import os
from pyknx import logger

__doc__ = __doc__.format(scriptname=os.path.basename(__file__))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=__doc__)
    args = parser.parse_args()

    # Configure logger.
    logger.initLogger(None)

    print(pyknx.version)



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

import sys
import re
import getopt
import time
from xml.dom.minidom import parseString, Document, Element
from threading import *
from pyknx import tcpsocket, logger

class Linknx:
    class SendMessageThread(Thread):
        def __init__(self, message, commandName, linknx):
            Thread.__init__(self, name='SendMessageThread (id={0})'.format(id(self)))
            self.socket = tcpsocket.Socket()
            self.linknx = linknx
            self.messageWithEncodingHeader = '<?xml version="1.0" encoding="utf-8"?>' + message
            self.commandName = commandName
            self.finalStatus = None
            self.answerDom = None
            self.error = None

        def run(self):
            try:
                self.socket.connect((self.linknx.host, self.linknx.port))
                logger.reportDebug('Message sent to linknx: ' + self.messageWithEncodingHeader)
                answer = self.socket.sendString(self.messageWithEncodingHeader, encoding='utf8')
                while True:
                    logger.reportDebug('Linknx answered ' + answer)
                    answerDom = parseString(answer[0:answer.rfind(chr(4))])
                    execNodes = answerDom.getElementsByTagName(self.commandName)
                    status = execNodes[0].getAttribute("status")
                    logger.reportDebug('Linknx Answer Status = {0}'.format(status))
                    if status == "ongoing":
                        # Wait for the final status.
                        answer = self.socket.waitForStringAnswer()
                        logger.reportDebug('New answer is {0}'.format(answer))
                    else:
                        if status != "success":
                            self.error = _getErrorFromXML(execNodes[0])
                            logger.reportError(error)
                        self.finalStatus = status
                        self.answerDom = answerDom
                        break
            finally:
                self.socket.close()
                logger.reportDebug('Thread is now stopped.')

        @property
        def isFinalized(self):
            return self.finalStatus != None

        def _getErrorFromXML(self, answer):
            errorMessage = ""
            for child in answer.childNodes:
                if child.nodeType == child.TEXT_NODE:
                    errorMessage += child.data
            return errorMessage

    """
    The wrapper of an instance of Linknx.

    This class eases access to Linknx functionalities: it can retrieve objects (see getObject() and getObjects()), or configuration.
    """

    class InvalidObjectIdException(Exception):
        """ The object id is not valid in the Linknx instance. """
        def __init__(self, objectId):
            self._objectId = objectId

        def __str__(self):
            return 'Object {0} does not exist.'.format(self._objectId)

        def __repr__(self):
            return 'InvalidObjectIdException({0})'.format(self._objectId)

    def __init__(self, hostname='localhost', port=1028):
        """ Initialize a Linknx wrapper. """
        self._host = hostname
        self._port = port
        self._config = None
        self._objectConfig = None
        self._objects = {}

    @property
    def host(self):
        return self._host

    @property
    def port(self):
        return self._port

    @property
    def address(self):
        return (self._host, self._port)

    @property
    def emailServerInfo(self):
        emailServerElements = self.config.getElementsByTagName('emailserver')
        if len(emailServerElements) == 0:
            return None
        else:
            serverElt = emailServerElements[0]
            if serverElt.getAttribute('type') != 'smtp': return None
            host = serverElt.getAttribute('host').split(':')
            if len(host) != 2: return None
            fromAddr = serverElt.getAttribute('from')
            return (host[0], int(host[1]), fromAddr)

    @property
    def config(self):
        """
        Get the raw XML configuration currently used by the running Linknx instance.

        See objectConfig property for an alternative of these raw data.

        """

        if self._config is None:
            xmlConfig = self.sendMessage("<read><config></config></read>", 'read').getElementsByTagName('read')[0]
            self._config = xmlConfig.getElementsByTagName('config')[0]

        return self._config

    @property
    def objectConfig(self):
        """ Return a dictionary of the configuration of each Linknx object. Keys are object ids, values are ObjectConfig instances. """
        if not self._objectConfig:
            objectsConfigNode = self.config.getElementsByTagName('objects')[0]
            self._objectConfig = {} # key is objectId, value is an ObjectConfig
            for objectConfigNode in objectsConfigNode.getElementsByTagName('object'):
                objectConfig = ObjectConfig(objectConfigNode)
                self._objectConfig[objectConfig.id] = ObjectConfig(objectConfigNode)

        return self._objectConfig

    def executeAction(self, actionDetails):
        if isinstance(actionDetails, str):
            actionXML = actionDetails
        elif isinstance(actionDetails, Document):
            actionXML = actionDetails.childNodes[0].toxml()
        elif isinstance(actionDetails, Element):
            actionXML = actionDetails.toxml()
        else:
            raise Exception('Unsupported action details: must be a minidom XML document or element or an XML string.')

        # Build XML document to send to linknx.
        self.sendMessage('<execute>{action}</execute>'.format(action=actionXML), 'execute', waitsForAnswer=False)
        logger.reportDebug('Action execution has been sent to linknx.')

    def waitForRemoteConnectionReady(self):
        """
        Wait for Linknx's XML server to accept incoming connections.

        This method should be called if unsure about when the Linknx has been started. It may take some time to get ready.
        This method attempts to connect to Linknx during 10 seconds and raises an Exception if Linknx still is unreachable after this delay.

        """
        # Ask for config.
        logger.reportInfo('Start connecting to linknx on {0}.'.format(self.address))
        attemptId = 0
        maxAttemptCount = 10
        while attemptId < maxAttemptCount:
            attemptId += 1
            try:
                conf = self.config

                # Linknx is ready if we reach this point.
                logger.reportInfo('Linknx is up and ready, let\'s start.')
                return
            except ConnectionRefusedError:
                logger.reportInfo('Linknx is not yet ready...  (attempt {0}/{1})'.format(attemptId, maxAttemptCount))

            except Exception as e:
                logger.reportException()

            time.sleep(1)

        raise Exception('Linknx is not reachable.')

    def getObject(self, id):
        """ Get the object of given identifier. """
        if id is None: return None

        obj = self._objects.get(id)
        if obj is None:
            obj = Object(id, self)
            self._objects[id] = obj
        return obj

    def getObjects(self, pattern='.*'):
        """ Get the objects whose identifiers match the given regex pattern. Default pattern matches all identifiers, thus calling getObjects() returns all objects. """
        regex = re.compile(pattern)
        objects = []
        for id in self.objectConfig.keys():
            if regex.search(id) is None: continue
            objects.append(self.getObject(id))
        return objects

    def sendMessage(self, message, commandName, waitsForAnswer=True):
        """
        Sends an XML message to Linknx.

        This function is implemented mainly for internal purposes. The end user is unlikely to call it directly.
        message -- An XML request that follows Linknx XML protocol.
        commandName -- The name of the XML command that is sent.
        waitsForAnswer -- If True, this method blocks until linknx has sent its final status. Otherwise, the method returns immediately. Linknx's answer would then be logged when it arrives.
        Returns an XML document that corresponds to Linknx answer if waitsForAnswer is False, None otherwise.

        """
        # logger.reportDebug('Sending message to linknx: ' + message)
        messagingThread = Linknx.SendMessageThread(message, commandName, self)

        if waitsForAnswer:
            # CAUTION - PERFORMANCE WARNING
            # Call run() directly to avoid starting Thread. That will let the
            # current thread naturally block until run() returns with linknx's answer.
            # Otherwise, we would have to wait for SendMessageThread completion
            # in an infinite while loop with a harcoded sleep time at the end of
            # each iteration. A sleep time of 0.1s is too long in most cases (it
            # significantly slows down the overall request time) and a sleep
            # time of, say, 0.01 may still be too long in some cases. Client
            # expects our method to return as soon as possible.
            # Thread's work implementation may have been extracted into another
            # object to make things cleaner but reusing the Thread object is
            # quite straightforward.
            messagingThread.run()

            if messagingThread.finalStatus != 'success':
                raise Exception(messagingThread.error)

            return messagingThread.answerDom
        else:
            # Start thread and leave.
            # Do not care about final status here. Error would be logged by
            # messaging thread. Client does not want to wait for answer so not
            # notifying the error with an exception makes sense.
            messagingThread.start()
            return None

class ObjectConfig:
    def __init__(self, configNode):
        """    Configuration of an object in Linknx. """

        self.id = configNode.getAttribute('id')
        self.xml = configNode
        self.type = configNode.getAttribute('type')
        self.gad = configNode.getAttribute('gad')
        self.callback = configNode.getAttribute('pyknxcallback')
        self.init = configNode.getAttribute('init') if configNode.hasAttribute('init') else 'request'
        self.flags = configNode.getAttribute('flags') if configNode.hasAttribute('flags') else 'cwtu'
        self.caption = ObjectConfig.getTextInElement(configNode, mustFind=False).strip('\n\t ')
        firstTypeDigit = self.type[0:self.type.find('.')]
        if firstTypeDigit == '1':
            self.typeCategory = 'bool'
        elif firstTypeDigit in ['5', '6', '7', '8', '9', '12', '13', '29']:
            if self.type in ['5.001', '5.003', '9.xxx']:
                self.typeCategory='float'
            else:
                self.typeCategory = 'int'
        elif firstTypeDigit == '14':
            self.typeCategory = 'float'
        elif firstTypeDigit in ('4', '16', '28'):
            self.typeCategory = 'string'
        else:
            logger.reportWarning('Object ' + self.id + ' has an unsupported type ' + self.type)
            self.typeCategory = 'unknown'

    @staticmethod
    def getTextInElement(elt, mustFind = True):
        text = ''
        for node in elt.childNodes:
            if node.nodeType == node.TEXT_NODE:
                # if text:
                    # raise Exception('Multiple text nodes in the same element are not supported.')
                text = text + node.data

        if mustFind and not text:
            raise Exception('Missing text in element {0}'.format(elt.nodeName))
        return text

class Object(object):

    """ Linknx object. """

    def __init__(self, id, linknx):
        """
        Initialize an object from Linknx.

        id -- Identifier of the object. Corresponds to the id attribute in XML configuration.
        linknx -- Linknx instance that provides the object.
        """
        self._id = id
        self._linknx = linknx
        if not id in linknx.objectConfig:
            raise Linknx.InvalidObjectIdException(id)

        self._objectConfig = linknx.objectConfig[id]

    @property
    def id(self):
        """ Object identifier. This is the id attribute in XML configuration. """
        return self._id

    @property
    def caption(self):
        """ Object caption. This is the text of the object element in the XML configuration. """
        return self._objectConfig.caption

    @property
    def gad(self):
        """ Group Address of the object. This is the gad attribute in XML configuration. """
        return self._objectConfig.gad

    @property
    def linknx(self):
        """ Return the Linknx instance that provides the object. """
        return self._linknx

    @property
    def xml(self):
        """ Return the xml element corresponding to this object in Linknx configuration. """
        return self._objectConfig.xml

    @property
    def type(self):
        """ Return the string corresponding to the type attribute in XML configuration. """
        self._objectConfig.type

    @property
    def value(self):
        """ Read object's value from linknx. """
        message='<read><objects><object id="' + self._id + '"/></objects></read>'

        answerDom = self._linknx.sendMessage(message, 'read')

        readNodes = answerDom.getElementsByTagName("read")
        valueStr = None
        objectValues = {}
        objectsNodes = readNodes[0].getElementsByTagName("objects")
        objectNodes = objectsNodes[0].getElementsByTagName("object")
        for objectNode in objectNodes:
            objectId = objectNode.getAttribute("id")
            objectValue = objectNode.getAttribute("value")
            objectValues[objectId] = objectValue
        valueStr = objectValues[self._id]

        if self._objectConfig.typeCategory == 'bool':
            return valueStr in ['on', '1', 'yes', 'true']
        elif self._objectConfig.typeCategory == 'int':
            return int(valueStr)
        elif self._objectConfig.typeCategory == 'float':
            return float(valueStr)
        else:
            return valueStr

    def convertValueToString(self, objValue):
        if self._objectConfig.typeCategory == 'bool':
            if isinstance(objValue, bool):
                objectValue = 'on' if objValue else 'off'
            else:
                objValueStr = str(objValue).lower()
                if objValueStr in ['true', 'on', 'yes', '1']:
                    objectValue = 'on'
                elif objValueStr in ['false', 'off', 'no', '0']:
                    objectValue = 'off'
                else:
                    raise Exception('For object {1}: Unable to convert {0} to boolean.'.format(objValue, self._id))
        elif self._objectConfig.typeCategory == 'int':
            objectValue = str(objValue)
        elif self._objectConfig.typeCategory == 'float':
            objectValue = str(objValue)
        elif self._objectConfig.typeCategory == 'string':
            objectValue = str(objValue)
        else:
            raise Exception('Unsupported type ' + self._objectConfig.typeCategory)

        return objectValue

    @value.setter
    def value(self, objValue):
        """ Write object's value to linknx. """
        # Convert value to the linknx format.
        logger.reportDebug('Attempting to set value of {0} to {1}'.format(self._id, objValue))
        objectValue = self.convertValueToString(objValue)

        if not objValue is objectValue:
            logger.reportDebug('Value has been converted to ' + str(objectValue))

        # Initialize DOM with a simple string, then use minidom to write
        # attributes so that special characters are properly encoded (for
        # instance, &ampersand; in place of &, etc).
        messageDom = parseString('<write><object/></write>')
        objectNode = messageDom.getElementsByTagName('object')[0]
        objectNode.setAttribute('id', self._id)
        objectNode.setAttribute('value', objectValue)
        answerDom = self._linknx.sendMessage(messageDom.toxml(), 'write')

    def __repr__(self):
        return self.id

    def __str__(self):
        return self.id



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

import socket
from pyknx import logger

class Socket:
    def __init__(self):
        self._socket = socket.socket()
        self._socket.settimeout(5)
        # self._socket.setblocking(1)

    def bind(self, address):
        self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self._socket.bind(address)
        self._socket.listen(5)

    def connect(self, address):
        self._socket.connect(address)

    def close(self):
        # self._socket.shutdown(socket.SHUT_RDWR)
        self._socket.close()

    def waitForString(self, encoding = 'utf8', endChar=chr(4)):
        responseBytes, connection = self.waitForData(endChar.encode(encoding))
        if connection is None:
            return (None, None)
        else:
            return (responseBytes.decode(encoding), connection)

    def waitForData(self, endChar = chr(4)):
        # Wait for a connection.
        try:
            conn, address = self._socket.accept()
        except socket.timeout:
            return (None, None)

        # Read incoming data.
        data = b''
        while True:
            try:
                chunk = conn.recv(4)
                endCharIx = chunk.find(endChar)
                isLastChunk = endCharIx != -1
                if isLastChunk:
                    chunk = chunk[:endCharIx]
                data += chunk
                if isLastChunk:
                    break
            except:
                logger.reportException('Exception when waiting for incoming data.')
                try:
                    if not conn is None: conn.close()
                except:
                    logger.reportException('Could not close connection. Connection is discarded and process continues.')
                    pass
                return None, None
        return (data, conn)

    def sendString(self, string, encoding = 'utf8', endChar = chr(4)):
        # Encode the passed string to get raw bytes ready for transfer.
        bytes = string.encode(encoding)

        # Send that over socket.
        responseBytes = self.sendData(bytes, endChar.encode(encoding))

        # Decode the response string from raw bytes.
        return responseBytes.decode(encoding)

    def sendData(self, data, endSequence):
        """
        Sends raw bytes to the other end.
        data: bytes object that contains the data to send.
        endSequence: bytes object that represents the end sequence when sending/receiving data to/from the socket.
        """
        self._socket.sendall(data + endSequence)
        self._socket.settimeout(70)
        answer = self.waitForAnswer(endSequence)
        return answer

    def waitForStringAnswer(self, encoding = 'utf8', endChar = chr(4)):
        # Send that over socket.
        responseBytes = self.waitForAnswer(endChar.encode(encoding))

        # Decode the response string from raw bytes.
        return responseBytes.decode(encoding)

    def waitForAnswer(self, endSequence):
        answer = b''
        while True:
            chunk = self._socket.recv(4096)
            answer += chunk
            if not chunk or chunk[len(chunk)-len(endSequence):] == endSequence:
                break
        return answer



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Starts an instance of the Pyknx communicator daemon.
The daemon is aimed at listening for linknx events for object changes in order to call Python functions that are implemented in a script
that is passed to the daemon. This process allows for execution of Python code in reaction of object changes on linknx's end.
"""

import sys
import argparse
import signal
import time
import logging
from pyknx import logger
from pyknx.linknx import *
from pyknx.communicator import *

def parseAddress(addrStr, option):
    ix = addrStr.find(':')
    if ix < 0:
        raise Exception('Malformed value for ' + option +'. Expecting a tuple (hostname:port)')
    return (addrStr[0:ix], int(addrStr[ix + 1:]))

def makeArgumentParser(description):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-c', '--comm-addr', dest='communicatorAddress', help='Address of the communicator. This argument must specify the hostname or the ip address followed by a colon and the port to listen on. Default is "localhost:1029"', default='localhost:1029')
    parser.add_argument('-l', '--linknx-addr', dest='linknxAddress', help='Address of the linknx server to bind to. This argument must specify the hostname or the ip address followed by a colon and the port. Default is "localhost:1028"', default='localhost:1028')
    parser.add_argument('userFile', help='use FILE as the user python script that implements callbacks functions declared in the linknx configuration (see the pyknxcallback attributes in XML).', metavar='FILE')
    parser.add_argument('--log-file', dest='logFile', help='write communicator\'s output to FILE rather than to standard output.', metavar='FILE', default=None)
    parser.add_argument('-d', '--daemonize', help='ask daemon to detach and run as a background daemon.', action='store_true', default=False)
    parser.add_argument('--pid-file', dest='pidFile', help='writes the PID of the daemon process to PIDFILE.', metavar='PIDFILE')
    parser.add_argument('-v', '--verbose', dest='verbosityLevel', help='set verbosity level. Default is "error".', metavar='LEVEL', choices=[l.lower() for l in logger.getLevelsToString()], default='error')
    return parser

if __name__ == '__main__':
    parser = makeArgumentParser(__doc__)
    args = parser.parse_args()

    # Configure logger.
    logger.initLogger(None, args.verbosityLevel.upper())

    args.communicatorAddress = parseAddress(args.communicatorAddress, 'communicator address')

    try:
        Communicator.run(args.linknxAddress, args.userFile, args.communicatorAddress, logFile=args.logFile, daemonizes=args.daemonize, pidFile=args.pidFile)
    except SystemExit:
        # This is a normal exit.
        pass
    except:
        logger.reportException()



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Logging module for pyknx.

This module is based on the standard logging module and adapted to better suit pyknx needs.
This module automatically adds a handler for signal USR1, which should be sent to notify the application that its log file has been moved and must be reloaded.
"""

import logging
import traceback
import os.path
import sys
import signal

logHandlers = []
stdOutLog = None # None to disable stdout logging, otherwise log level.
fileLog = None # Tuple (filename, level)

def getLevelsToString():
    return ('ERROR', 'WARNING', 'INFO', 'DEBUG')

def initLogger(fileLogInfo=None, stdoutLogLevel=logging.INFO, usesDetailedLogging=True):
    """
    Initialize the logging system. Should be called prior to any other function of this module.

    fileLogInfo -- A tuple that contains the log filename and log verbosity. Should be None to deactivate file logging.
    stdoutLogLevel -- Verbosity level to use when writing log to stdout. Should be None to deactivate logging to stdout.

    """
    global logHandler
    global logFilename
    global logLevel

    _setHandlers(None, None)
    _setHandlers(fileLogInfo, stdoutLogLevel, usesDetailedLogging)
    logging.getLogger().setLevel(logging.DEBUG)
    signal.signal(signal.SIGUSR1, _usr1SignalHandler)
    reportDebug('Logger initialized with fileLogInfo={fli} stdoutLogLevel={stdoutLL} usesDetailedLogging={usesDetailedLogging}.'.format(fli=fileLogInfo, stdoutLL=stdoutLogLevel, usesDetailedLogging=usesDetailedLogging))

def parseLevel(levelToString):
    """    Parses a string that represents the log level. The string should be the same than the level in the logging module. """
    lValue = levelToString.lower()
    if lValue == 'error':
        return logging.ERROR
    elif lValue == 'warning':
        return logging.WARNING
    elif lValue == 'info':
        return logging.INFO
    elif lValue == 'debug':
        return logging.DEBUG
    else:
        raise Exception('Unknown verbosity level ' + levelToString)

def _setHandlers(fileLogInfo, stdoutLogLevel, usesDetailedLogging=True):
    global logHandlers
    global fileLog
    global stdOutLogLevel

    logger = logging.getLogger()

    # Remove previous handlers.
    for handler in logHandlers:
        logger.removeHandler(handler)
        if isinstance(handler, logging.FileHandler):
            handler.close()

    if not fileLogInfo is None and not isinstance(fileLogInfo, tuple):
        raise Exception('File log info should be specified with a tuple (filename, loglevel).')

    fileLog = fileLogInfo
    if(isinstance(fileLog, tuple) and isinstance(fileLog[1], str)):
            fileLog = (fileLog[0], parseLevel(fileLog[1]))
    stdOutLog = stdoutLogLevel
    if(isinstance(stdOutLog, str)):
        stdOutLog = parseLevel(stdOutLog)

    # Create new handlers.
    logHandlers = []
    if not fileLog is None:
        dir = os.path.normpath(os.path.dirname(fileLog[0]))
        if not os.path.isdir(dir):
            os.makedirs(dir)

        _addHandler(logging.FileHandler(fileLog[0]), fileLog[1], usesDetailedLogging)
    if not stdOutLog is None:
        _addHandler(logging.StreamHandler(), stdOutLog, usesDetailedLogging)

def _addHandler(handler, logLevel, usesDetailedLogging=True):
    global logHandlers

    handler.setLevel(logLevel)
    logHandlers.append(handler)
    if usesDetailedLogging:
        formatter = logging.Formatter('%(asctime)s [%(levelname)s] [%(threadName)s] [%(callerfilename)s:%(callerlineno)d] %(message)s')
    else:
        formatter = logging.Formatter('%(message)s')
    handler.setFormatter(formatter)
    logging.getLogger().addHandler(handler)

def _usr1SignalHandler(signalNumber, frame):
    if signalNumber == signal.SIGUSR1:
        reportInfo('USR1 signal caught. Means that log file has to be reloaded.')
        _setHandlers(fileLog, stdOutLog)

def _reportMessage(message, level):
    stack = traceback.extract_stack()
    frame = stack[len(stack) -3]
    extraDict={'callerfilename' : os.path.basename(frame[0]), 'callerlineno' : frame[1]}
    logging.getLogger().log(level, message, extra=extraDict)

def reportDebug(message):
    """ Reports a debug message. """
    _reportMessage(message, logging.DEBUG)

def reportError(message):
    """ Reports an error message. """
    _reportMessage(message, logging.ERROR)
    # logging.getLogger().error(message)

def reportWarning(message):
    """ Reports a warning message. """
    _reportMessage(message, logging.WARNING)
    # logging.getLogger().warning(message)

def reportInfo(message):
    """ Reports an informational message. """
    _reportMessage(message, logging.INFO)
    # logging.getLogger().info(message)

def reportException(message=None):
    """ Reports an exception. Exception info is gotten from sys.exc_info(). """
    if not message: message = 'Exception caught.'
    _reportMessage(message + ' Traceback is:\n' + traceback.format_exc(), logging.ERROR)



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Pyknx is a package that is aimed at providing basic functionality related to communicating with a Linknx instance. It should help in sending or receiving data to/from Linknx.

linknx.py: common module that implements the communication with a linknx server. With this module, one can retrieve linknx objects, read or write their value, read linknx configuration, ...
communicator.py: this module contains the Communicator daemon, whose purpose is to receive events from linknx, through ioports.  It is then easy to write callbacks to react to object modifications. Additional scripts based on pyknx are provided (see below) in order to make this bidirectional communication with linknx just a few keystrokes away from now!
logger.py: internal module that provides logging functionality for the package.
tcpsocket.py: an internal module that implements common functionality related to socket communication. The end-user is unlikely to use this module directly.
"""
__all__ = ['linknx', 'communicator']

class Version:
    def __init__(self, major, minor, revision, maturityLevel, maturationStep):
        if maturityLevel not in ('a', 'b', ''):
            raise Exception('Unknown maturity level.')
        self.major = major
        self.minor = minor
        self.revision = revision
        self.maturityLevel = maturityLevel
        self.maturationStep = maturationStep

    def __repr__(self):
        return '{major}.{minor}.{revision}{maturityLevel}{maturationStep}'.format(**vars(self))

version = Version(2, 0, 0, 'b', 11)
__version__=str(version)




#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

import sys
import getopt
import time
import os.path
import logging
import importlib
import signal
from threading import *
from pyknx import tcpsocket, logger
from pyknx.linknx import *

class CallbackContext(object):

    """
    The context of the call of one of the user-defined functions, either from Linknx or from pyknxcall.py.

    This class always exposes the communicator and Linknx instances.
    When call is issued as a consequence of an object value change in Linknx, the context also provides the object that changed.
    When call is issued from pyknxcall.py script all extra arguments provided on the command line are accessible through homonym properties. For instance, when executing 'pyknxcall.py -amyArg=foo myFunction'
    the function myFunction is passed an instance of CallbackContext which contains a member named 'myArg' and whose value is 'foo'.

    """

    def __init__(self, communicator, args={}):
        # if not args is None and args.has_key('objectId'):
            # self._object = linknx.getObject(args['objectId'])
        # else:
            # self._object = None
        self._communicator = communicator
        self._args = args

        # Create members for easy access.
        if args != None:
            exp = re.compile('[^0-9a-zA-Z_]')
            for argName, argValue in args.items():
                # Rename args to comply with Python naming rules.
                changedName = re.sub(exp, '_', argName)
                if changedName != argName:
                    logger.reportWarning('Argument {0} renamed to {1} to comply with naming rules.'.format(argName, changedName))

                # Create argument.
                vars(self)[changedName] = argValue

        if 'objectId' in vars(self):
            self._object = self.linknx.getObject(self.objectId)
        else:
            self._object = None

    @property
    def object(self):
        """ A wrapper of the Linknx object that is related to the event. """
        return self._object

    @property
    def linknx(self):
        """ The instance of the Linknx wrapper the communicator is linked with. """
        return self._communicator.linknx

    @property
    def communicator(self):
        """ The instance of the pyknx communicator that receives events either from Linknx or from the pyknxcall.py utility script. """
        return self._communicator

    @property
    def customArgs(self):
        """
        A dictionary that contains all arguments of the user-defined function call.

        Keys are the argument names, values their corresponding argument value.
        """
        return self._args

    def getArgument(self, argName, defaultValue = None):
        """
        A convenience method that allows to get the value of a given argument.

        This method is equivalent to searching in the customArgs dictionary.
        argName -- The name of the argument to get.
        defaultValue -- The value to return if no such argument is found in this context.

        """

        if argName in self.customArgs:
            return self.customArgs[argName]
        else:
            return defaultValue

    def __str__(self):
        return str(self._args)

class Communicator:

    """
    The communicator daemon that is intended to be executed once Linknx is started.

    This daemon continuously listens on its dedicated port for incoming data either from Linknx or from the pyknxcall.py utility script.
    It is aimed at performing calls of functions in the user-defined file specified in the initializer of the class.

    """
    class Listener(Thread):
        """ Thread that listens for incoming connection from linknx. """
        def __init__(self, address, communicator):
            Thread.__init__(self, name='Communicator Listening Thread')
            self._address = address
            self._isStopRequested = False
            self._socket = tcpsocket.Socket() # socket to listen on for information coming from linknx.
            self._communicator = communicator
            self.linknx = self._communicator.linknx
            self.isReady = False

        def isListening(self):
            return not self._socket is None and self.isReady
        @property
        def isStopped(self):
            return self._socket is None

        def run(self):
            logger.reportInfo('Listening on ' + str(self._address))
            self._isStopRequested = False
            try:
                self._socket.bind(self._address)

                # Thread loop.
                while not self._isStopRequested:
                    self.isReady = True
                    data, conn = self._socket.waitForString(endChar='$')
                    # Throw data away if script has not been initialized yet.
                    # See startListening for details.
                    if data is None or not self._communicator.isUserScriptInitialized:
                        time.sleep(0.1)
                        continue

                    logger.reportDebug('Data received: {0}'.format(data))

                    # Handle request.
                    tokens = data.split('|')
                    callbackName = tokens[0]
                    # Parse arguments. First is object id.
                    args={}
                    for token in tokens[1:]:
                        argName, sep, argValue = token.partition('=')
                        if argValue: argValue = argValue.strip()
                        args[argName.strip()] = argValue
                    context = CallbackContext(self, args)
                    res = self._communicator._executeUserCallback(callbackName, context)
                    if res:
                        conn.sendall(res + '$')
                    conn.close()
            except Exception as e:
                logger.reportException()
            finally:
                logger.reportDebug('Closing socket...')
                self._socket.close()
                logger.reportInfo('Socket closed. Listening terminated.')
                self._socket = None

        def stop(self):
            logger.reportInfo('Stopping listener thread...')
            self._isStopRequested = True

    def __init__(self, linknx, userFile, address=('localhost',1029), userScriptArgs={}):

        """
        Initialize the daemon.

        linknx -- The wrapper of the Linknx server to communicate with.
        userFile -- The file that implements the user-defined functions to be called when objects which have a pyknxcallback attribute in the Linknx configuration change.
        address -- The address the communicator will listen on, defined as a tuple (ip address, port). The default is ('localhost', 1029).
        userScriptArgs -- A dictionary of extra arguments to expose in the CallbackContext instance passed to the initializeUserScript function (if implemented in the user file). It defaults to empty.

        """
        self._address = address
        self._listenerThread = None # thread that owns the listening socket, None if not listening.
        self._userFile = userFile
        self._linknx = linknx
        self._userModule = None
        self._userScriptArgs = userScriptArgs
        self.isUserScriptInitialized = False

    @property
    def isListening(self):
        """ Tell whether the communicator is currently listening for incoming data. """
        return not self._listenerThread is None

    @property
    def linknx(self):
        """ Return the wrapper of the Linknx instance the communicator is working with. """
        return self._linknx

    @property
    def address(self):
        """ Return the listening address as a tuple (ip address, port). """
        return self._address

    def _loadUserFile(self):
        # Append the directory that contains the user script to python path.
        if self._userFile:
            dirName, fileName = os.path.split(self._userFile)
            dirName = os.path.abspath(dirName)
            moduleName, fileExt = os.path.splitext(fileName)
            sys.path.append(dirName)
            logger.reportDebug('_loadUserFile: moduleName={0} fileExt={1} dirName={2}'.format(moduleName, fileExt, dirName))
            self._userModule = importlib.import_module(moduleName)
            logger.reportDebug('Imported {0}'.format(self._userModule.__file__))
            return True
        else:
            logger.reportError('No user file specified.')
            return False

    @staticmethod
    def run(linknxAddress, userFile, communicatorAddress, userScriptArgs=None, verbosityLevel=logging.INFO, logFile=None, daemonizes=False, pidFile=None):
        def signal_handler(signal, frame):
            logger.reportInfo('Terminating...')
            communicator.stopListening()

        # Init logger.
        if not logFile is None:
            logger.initLogger((logFile, verbosityLevel), None)
        else:
            logger.initLogger(None, verbosityLevel)

        if isinstance(linknxAddress, tuple):
            linknxAddr = (linknxAddress[0], int(linknxAddress[1]))
        elif isinstance(linknxAddress, str):
            tokens = linknxAddress.split(':')
            linknxAddr = (tokens[0], int(tokens[1]))
        else:
                raise Exception('Unrecognized linknx address format.')
        linknx = Linknx(linknxAddr[0], int(linknxAddr[1]))

        # Fork if requested.
        if daemonizes:
            pid = os.fork()
        else:
            pid = os.getpid()

        if pid != 0 and pidFile != None:
            with open(pidFile, 'w') as f:
                f.write(str(pid))

        # If we are in the parent process (when forking), there is nothing left
        # to do. The communicator should run in the child process.
        if pid != 0 and daemonizes:
            return

        # Start communicator.
        communicator = Communicator(linknx, userFile, communicatorAddress, userScriptArgs=userScriptArgs)
        communicator.startListening()

        signal.signal(signal.SIGINT, signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)

        # Main loop.
        while communicator.isListening:
            time.sleep(1)

        # Clean pid file.
        if pidFile != None and os.path.exists(pidFile):
            os.remove(pidFile)

    def startListening(self):

        """
        Start the communicator. It is then waiting for incoming information from Linknx.

        If provided, the initializeUserScript function of the user file is called. Its context contains the arguments that were optionally given to the __init__ method.

        """
        if self.isListening: return

        # Make sure linknx is ready.
        self.linknx.waitForRemoteConnectionReady()

        # Start listening early to avoid communication errors from linknx. Those
        # errors are never harmful but the user may be surprized and worried
        # about them! 
        self._listenerThread = Communicator.Listener(self._address, self)
        self._listenerThread.start()
        timeout = time.time() + 4
        while not self._listenerThread.isReady and time.time() < timeout:
            time.sleep(0.3)
        if not self._listenerThread.isReady:
            raise Exception('Could not initialize listening socket.')

        # Initialize user-provided script. The purpose of this callback is to
        # let the user initialize its script by reading state from linknx (and
        # possibly anywhere else). Thus, linknx should not raise events yet,
        # since user script would likely be partially initialized. The
        # isUserScriptInitialized flag is used for that purpose.
        if self._loadUserFile():
            logger.reportInfo('Initializing user script...')
            try:
                self._executeUserCallback('initializeUserScript', CallbackContext(self, args=self._userScriptArgs), True)
            except Exception as e:
                logger.reportException('User script initialization failed, communicator will stop immediately.')
                self.stopListening()
                return
            logger.reportInfo('User script initialized.')
        self.isUserScriptInitialized = True


    def stopListening(self):
        """ Stop communicator. No new incoming connection will be possible. """
        # Notify user script first. This allows linknx to notify a few object
        # changes before communicator really stop listening.
        if self._userFile and self.isUserScriptInitialized:
            self._executeUserCallback('finalizeUserScript', CallbackContext(self), True)
            logger.reportInfo('User script finalized.')

        if not self.isListening: return
        self._listenerThread.stop()

        # Wait for listener thread to end (to be sure that no callback
        # request originating from linknx can reach the user script anymore).
        while not self._listenerThread.isStopped:
            time.sleep(0.5)
        self._listenerThread = None

        if self._userFile:
            self._executeUserCallback('endUserScript', CallbackContext(self), True)
            logger.reportInfo('User script ended.')

    def _executeUserCallback(self, callbackName, context, isOptional=False):
        try:
            if hasattr(self._userModule, callbackName):
                logger.reportDebug('Calling user callback {0} with context {1}'.format(callbackName, context))
                callback = getattr(self._userModule, callbackName)
                res = callback(context)
                logger.reportDebug('Callback {0} returned {1}'.format(callbackName, res))
                return res
            else:
                message='No function {0} defined in {1}'.format(callbackName, self._userFile)
                if isOptional:
                    logger.reportInfo(message + ', skipping')
                else:
                    logger.reportWarning(message)
        except Exception as e:
            logger.reportException('User code execution failed.')



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Module that implements functionality common to all client scripts (pyknxread, pyknxwrite, pyknxexecute).
"""

import sys
import argparse
import traceback
import logging
from xml.dom.minidom import parseString
from threading import *
from pyknx import logger
from pyknx.linknx import *

def handleRequest(requestType, doc):
    logger.initLogger(None, logging.INFO, usesDetailedLogging=False)

    parser = argparse.ArgumentParser(description=doc)
    parser.add_argument('-s', '--server', dest='host', help='Hostname of the machine running the linknx daemon. Default is localhost.', default='localhost')
    parser.add_argument('-p', '--port', dest='port', help='Port linknx listens on. Default is 1028.', default=1028)
    if requestType == 'read':
        parser.add_argument('objectIds', help='ID represents the identifier of the object to read.', metavar='ID', nargs='+')
        parser.add_argument('-R', '--regex', action='store_true', help='ID in the "object" argument is interpreted as a regex and used to find objects to read. The pattern must comply with the \'re\' python module.')
        parser.add_argument('--value-only', action='store_true', help='Output the value of the queried object but do not prefix it with the object\'s id.')
        parser.add_argument('--expected-value', help='Expected value of the object. This script will exit with a non-zero return code if the value is not the expected one. This is useful when using this script in a "if" test of a shell script.')
    elif requestType == 'write':
        parser.add_argument('object', help='ID represents the identifier of the object to write to.', metavar='ID')
        parser.add_argument('value', help='Assigns VALUE to the object identified by ID.', metavar='VALUE')
    elif requestType == 'execute':
        parser.add_argument('--action', help='use the ACTION string as the XML representation of the action to execute rather than reading it from standard input.', metavar='ACTION')
    else:
        raise Exception('Unsupported request type "{0}".'.format(requestType))
    parser.add_argument('-v', '--verbose', dest='verbosityLevel', help='Set verbosity level. Default is "error".', metavar='LEVEL', choices=[l.lower() for l in logger.getLevelsToString()], default='error')
    args = parser.parse_args()

    # Configure logger.
    logger.initLogger(None, args.verbosityLevel.upper())

    # Start linknx.
    linknx = Linknx(args.host, int(args.port))
    try:
        if requestType == 'read':
            report = {}
            for objectId in args.objectIds:
                if not args.regex:
                    report[objectId] = linknx.getObject(objectId).value
                else:
                    for obj in linknx.getObjects(objectId):
                        report[obj.id] = obj.value

            # No object.
            if not report:
                logger.reportWarning('No object of given id.')
                sys.exit(10)

            # Count tabs to align columns.
            longestId = max([len(id) for id in report.keys()])
            succeeds = True
            for o, v in report.items():
                spaceCount = longestId - len(o)
                spaces=''
                while spaceCount > 0:
                    spaces+=' '
                    spaceCount -= 1
                if args.value_only:
                    print('{0}'.format(v))
                else:
                    print('{0} {2} {1}'.format(o, v, spaces))

                if args.expected_value != None:
                    obj = linknx.getObject(o)
                    convertedExpectedValue = obj.convertValueToString(args.expected_value)
                    convertedObjectValue = obj.convertValueToString(v)
                    succeeds = succeeds and convertedExpectedValue == convertedObjectValue

            if not succeeds: exit(100)

        elif requestType == 'write':
            linknx.getObject(args.object).value = args.value
        elif requestType == 'execute':
            if args.action == None:
                action = ''.join(sys.stdin.readlines())
            else:
                action = args.action
            linknx.executeAction(action)
        else:
            raise Exception('Unsupported request type.')

    except Exception as e:
        logger.reportError(sys.exc_info()[1])
        sys.exit(3)



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Let the communicator call a given function of the user script.

USAGE:
pyknxcall.py [-c communicatoraddress] [-a|--argument argname=argvalue [-a|--argument argname=argvalue [...]]] [-v level] functionName

OPTIONS:
    -c --comm-addr              Address of the communicator. This must be of the form <hostname:port> of <ipaddress:port>. Default is localhost:1029
    -a --argument               Argument to pass to user function. It is of the form argument_name=argument_value. Be careful not to insert whitespaces around the "=" sign.
    -v --verbose                Level of verbosity. Value must be one of the logging module (error, warning, info, debug)
    --help                      Display this help message and exit.

EXAMPLE:
    pyknxcall.py -a eventId=3 onEventStarted

    => calls the function onEventStarted in the user file with the -f argument of pyknxcommunicator.py. The implementation may look like this:
    def onEventStarted(context):
        print 'Event #{0} started.'.format(context.eventId)
"""

import sys
import getopt
import logging
from pyknx import logger, tcpsocket

def printUsage():
    print(__doc__)

def parseAddress(addrStr, option):
    ix = addrStr.find(':')
    if ix < 0:
        raise Exception('Malformed value for ' + option +'. Expecting a tuple (hostname:port)')
    return (addrStr[0:ix], int(addrStr[ix + 1:]))

if __name__ == '__main__':
    logger.initLogger(None, logging.INFO, usesDetailedLogging=False)
    try:
        options, remainder = getopt.getopt(sys.argv[1:], 'c:a:v:', ['comm-addr=', 'argument=', 'verbose=', 'help'])
    except getopt.GetoptError as err:
        logger.reportException()
        sys.exit(2)

    # Parse command line arguments.
    communicatorAddress = ('127.0.0.1',1029)
    arguments = []
    verbosity = logging.INFO
    for option, value in options:
        if option == '-c' or option == '--comm-addr':
            communicatorAddress = parseAddress(value, option)
        elif option == '-a' or option == '--argument':
            arguments.append(value)
        elif option == '--help':
            printUsage()
            sys.exit(1)
        elif option == '-v' or option == '--verbose':
            verbosity = logger.parseLevel(value)
        else:
            logger.reportError('Unrecognized option ' + option)
            sys.exit(2)

    if not remainder:
        logger.reportError('Missing function name.')
        printUsage()
        sys.exit(3)

    if len(remainder) > 1:
        logger.reportError('Too many arguments: {0}'.format(remainder))
        printUsage()
        sys.exit(4)

    functionName = remainder[0]

    # Init logger.
    logger.initLogger(None, verbosity, usesDetailedLogging=False)

    s = tcpsocket.Socket()
    s.connect(communicatorAddress)
    message=functionName
    for arg in arguments:
        message += '|{0}'.format(arg)
    s.sendString(message, endChar='$')




#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

from xml.dom.minidom import parse, parseString
from pyknx.linknx import ObjectConfig
from pyknx import logger
import sys
import getopt
import codecs
import logging
import socket

class Configurator:
    """ Object able to automatically patch the linknx configuration xml to add python bindings. """
    def __init__(self, sourceFile, outputFile, address, communicatorName='pyknxcommunicator'):
        self._address = address
        self._sourceFile = sourceFile
        self._outputFile = outputFile
        self._communicatorName = communicatorName
        self._config = None

    @property
    def config(self):
        if not self._config:
            if self._sourceFile != None:
                doc = parse(self._sourceFile)
            else:
                # Read from standard input.
                doc = parseString(self.readFileFromStdIn())
            self._config = doc.getElementsByTagName('config')[0]

        return self._config

    def readFileFromStdIn(self):
        return ''.join(sys.stdin.readlines())

    def cleanConfig(self):
        # Delete all pyknx rules before creating only those that apply to the
        # current config.
        rulesNode = self._getOrAddConfigElement(self.config, 'rules')
        prefixLength = len(self._communicatorName)
        configuredAtLeastOne = False
        for ruleNode in rulesNode.getElementsByTagName('rule'):
            ruleId = ruleNode.getAttribute('id')
            if ruleId[:prefixLength] == self._communicatorName:
                configuredAtLeastOne = True
                logger.reportInfo('Clean rule ' + ruleId + ' coming from a previous configure.')
                rulesNode.removeChild(ruleNode)

        if not configuredAtLeastOne:
            logger.reportInfo('Input XML config does not define any pyknx rule. Nothing to clean.')

        servicesNode = self._getOrAddConfigElement(self.config, 'services')
        ioportsNode = self._getOrAddConfigElement(servicesNode, 'ioports')
        for ioportNode in ioportsNode.getElementsByTagName('ioport'):
            if ioportNode.getAttribute('id') == 'pyknxcommunicator':
                logger.reportInfo('Clean ' + ioportNode.toxml())
                ioportsNode.removeChild(ioportNode)

    def createActionNode(self, callbackName, args):
        doc = self.config.ownerDocument
        actionNode = doc.createElement('action')
        actionNode.setAttribute('type', 'ioport-tx')
        actionNode.setAttribute('ioport', self._communicatorName)
        dataStr = callbackName
        if not args is None:
            for argName, argValue in args.items():
                dataStr += '|{0}={1}'.format(argName, argValue)
        actionNode.setAttribute('data', dataStr + '$')
        return actionNode

    def generateConfig(self):
        # Read xml to get pyknx special attributes.
        config = self.config
        doc = config.ownerDocument
        rulesNode = self._getOrAddConfigElement(config, 'rules')

        # Generate a rule for each object that has a callback in the user file.
        objectNodes = config.getElementsByTagName('objects')[0]
        configuredAtLeastOne = False
        for objectNode in objectNodes.getElementsByTagName('object'):
            objectConfig = ObjectConfig(objectNode)
            objectId = objectConfig.id
            if not objectConfig.callback:
                logger.reportDebug('No callback found for object ' + objectConfig.id)
                continue

            configuredAtLeastOne = True
            ruleNode = doc.createElement('rule')
            ruleId = '{0}{1}'.format(self._communicatorName, objectId)
            logger.reportInfo('Generating rule {0}'.format(ruleId))
            ruleNode.setAttribute('id', ruleId)
            ruleNode.setAttribute('init', 'false')
            conditionNode = doc.createElement('condition')
            conditionNode.setAttribute('type', 'object')
            conditionNode.setAttribute('id', objectId)
            # conditionNode.setAttribute('value', objectConfig.defaultValue)
            conditionNode.setAttribute('trigger', 'true')
            ruleNode.appendChild(conditionNode)
            actionListNode = doc.createElement('actionlist')
            actionListNode.setAttribute('type', 'if-true')
            ruleNode.appendChild(actionListNode)
            actionNode = self.createActionNode(objectConfig.callback, {'objectId' : objectId})
            actionListNode.appendChild(actionNode)
            # actionListIfFalseNode = actionListNode.cloneNode(True)
            # actionListIfFalseNode.setAttribute('type', 'on-false')
            # # ruleNode.appendChild(actionListIfFalseNode)
            rulesNode.appendChild(ruleNode)

        if not configuredAtLeastOne:
            logger.reportInfo('Nothing to do. None of the objects does define a pyknxcallback attribute.')
        else:
            # Add an ioport service for the communicator.
            servicesNode = self._getOrAddConfigElement(config, 'services')
            ioportsNode = self._getOrAddConfigElement(servicesNode, 'ioports')
            ioportNode = doc.createElement('ioport')
            ioportNode.setAttribute('id', self._communicatorName)
            try:
                hostIP = socket.gethostbyname(self._address[0])
            except:
                logger.reportWarning('Could not check that {0} is a valid ip address. Please check the output configuration. Linknx does not support hostnames, it requires IP address.'.format(self._address[0]))
                hostIP = self._address[0]
            ioportNode.setAttribute('host', hostIP) #gethostbyname converts the hostname into an ip. Linknx does not support ioport hostnames.
            ioportNode.setAttribute('port', str(self._address[1]))
            ioportNode.setAttribute('type', 'tcp')
            ioportsNode.appendChild(ioportNode)


    def writeConfig(self):
        if self._outputFile != None:
            outputXMLFile = codecs.open(self._outputFile, mode='w', encoding='utf-8')
            outputXMLFile.write(self.config.toxml())
            outputXMLFile.close()
            logger.reportInfo('Output config written to ' + self._outputFile)
        else:
            print(self.config.toxml())


    def _getOrAddConfigElement(self, parent, elementTagName):
        elementNodes = parent.getElementsByTagName(elementTagName)
        if not elementNodes:
            elementNode = parent.ownerDocument.createElement(elementTagName)
            parent.appendChild(elementNode)
            logger.reportInfo('No <' + elementTagName + '> element in config, creating one.')
        else:
            elementNode = elementNodes[0]
        return elementNode



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Modifies an XML config for Linknx so that it allows for communication with an instance of pyknxcommunicator.py
This script adds an ioport and a rule for each object that holds a pyknxcallback attribute in the source XML.
It always starts by cleaning any autogenerated rule from the input file and generates new ones unless the --clean option is passed.
Source XML configuration is read from standard input unless the -i option is set.
"""

from pyknx.configurator import Configurator
import argparse
import getopt
import sys
import logging
from pyknx import logger

def parseAddress(addrStr, option):
    ix = addrStr.find(':')
    if ix < 0:
        raise Exception('Malformed value for ' + option +'. Expecting a tuple (hostname:port)')
    return (addrStr[0:ix], addrStr[ix + 1:])

def makeArgumentParser(description):
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-i', '--input-file', dest='linknxConfig', help='use LKNCONF as the source linknx configuration rather than reading from standard input.', metavar='LKNCONF')
    parser.add_argument('-o', '--output-file', dest='outputFile', help='write the modified linknx configuration to FILE rather than to standard output.', metavar='FILE')
    parser.add_argument('-c', '--comm-addr', dest='communicatorAddress', help='Address of the communicator. This argument must specify the hostname or the ip address followed by a colon and the port to listen on. Default is "localhost:1029"', default='localhost:1029')
    parser.add_argument('-n', '--comm-name', dest='communicatorName', help='Name of the communicator. Used to prefix name of rules generated by this script and to name the ioport service as well. Default is "pyknxcommunicator"', default='pyknxcommunicator')
    parser.add_argument('--clean', help='Clean rules that were generated by this script but do not generate new rules.', action='store_true')
    parser.add_argument('-v', '--verbose', dest='verbosityLevel', help='set verbosity level. Default is "error".', metavar='LEVEL', choices=[l.lower() for l in logger.getLevelsToString()], default='error')
    return parser

if __name__ == '__main__':
    parser = makeArgumentParser(__doc__)
    args = parser.parse_args()

    # Configure logger.
    logger.initLogger(None, args.verbosityLevel.upper())

    args.communicatorAddress = parseAddress(args.communicatorAddress, 'communicator address')

    # Start configurator.
    configurator = Configurator(args.linknxConfig, args.outputFile, args.communicatorAddress, args.communicatorName)

    # Generate config.
    configurator.cleanConfig()
    if not args.clean:
        configurator.generateConfig()
    configurator.writeConfig()



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

from distutils.core import setup
import sys
import pyknx

if sys.version_info.major < 3:
    print('This package is compatible with Python 3 and above.')
    exit(4)

setup(name='pyknx',
      version=pyknx.__version__,
      description='Python bindings for Linknx',
      long_description=''.join(open('README.md').readlines()),
      author='Cyrille Defranoux',
      author_email='knx@aminate.net',
      maintainer='Cyrille Defranoux',
      maintainer_email='knx@aminate.net',
      license='GNU Public General License',
      url='https://github.com/2franix/pyknx/',
      packages=['pyknx'],
      data_files=[('.', ['README.md'])],
      scripts=['pyknxcommunicator.py', 'pyknxcall.py', 'pyknxread.py', 'pyknxwrite.py', 'pyknxexecute.py', 'pyknxclient.py', 'pyknxconf.py', 'pyknxversion.py'])



#!/usr/bin/python3

# Copyright (C) 2012-2014 Cyrille Defranoux
#
# This file is part of Pyknx.
#
# Pyknx is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Pyknx is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Pyknx. If not, see <http://www.gnu.org/licenses/>.
#
# For any question, feature requests or bug reports, feel free to contact me at:
# knx at aminate dot net

"""
Lightweight command line client for linknx. It is aimed at executing actions defined by a XML string. The action to execute is read from standard input unless the --action option is set.
Example of a valid action: <action type="set-value" id="kitchen_heating" value="comfort"/>. This string is passed as-is to linknx so refer to linknx\'s documentation for further details.
"""

from pyknx import client

if __name__ == '__main__':
    client.handleRequest('execute', __doc__)
